#include <ctype.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <limits.h>

#define todigit(a) ((a) - '0')

#ifdef __cplusplus
extern "C"
#endif

/**
 * Convert a string to an int (the base is 10).
 *
 * This function converts the prefix made of digits (with possibly '-'
 * character at the beginning) to an integer.  On success, the pointer at the
 * first non-digit character is returned in endp, on error NULL is stored to
 * endp and 0 is returned.  When the string is convertible, but the resulting
 * number would not fit into the 'int' type, it is considered as an error.
 *
 * No whitespaces before the first digit (or '-') are allowed.  The number can
 * not be prefixed with zeroes, that is 123 is a valid input, but 0123 or -0123
 * is not. In other words, the first digit in the number must not be zero.
 *
 * If the string starts with a character that is not a digit or '-', then
 * the function fails and returns NULL. If the string starts with '-' but is
 * not followed by a digit, NULL is returned.
 * NULL is also returned if the input string is empty.
 *
 * Example of convertible strings: "123", "-123", "-2314bde", "234aef"
 * Example of non-convertible strings: "k23", "+234", "+-f3w", " 234", " - 234".
 */
int strtoi(const char *str, const char **endp)
{
	const char *end;
	bool negative = 0;

	if (*str == '-') {
		negative = 1;
		++str;
	}

	end = str;

	/* move to the end of digits */
	while(*end && isdigit(*end++));

	*endp = end;

	/* end points either to the terminating zero
	 * or to the first non-digit character, so move to
	 * the first digit */
	--end;

	unsigned int pos = 1;
	int result = 0;
	int tmp;
	while (end != str) {
		assert(isdigit(*end));
		tmp = pos * (todigit(*end));
		/* check for overflow */
		assert(tmp >= 0);
		if (INT_MAX > result + tmp) {
			result += tmp;
		} else {
			/* the number is too big to fit to int */
			goto err;
		}

		--end;
		pos *= 10;
		assert(end >= str);
	}

	/* convert the last one */
	assert(end == str);
	assert(pos <= 1000000000 && pos % 10 == 0);

	/* this means numbers >= 3000000000, those do not
	 * fit into an integer */
	if (pos == 1000000000 && todigit(*end > 2))
		goto err;

	tmp = pos * (todigit(*end));
	assert(tmp >= 0);
	if (INT_MAX > tmp - (negative ? 1 : 0)) {
		if (negative && (INT_MAX - tmp == result - 1))
			return INT_MIN;
		else if (INT_MAX - tmp >= result - (negative ? 1 : 0))
			result += tmp;
		else
			goto err;
	} else {
		/* the number is too big to fit to int */
		goto err;
	}

	if (negative)
		return -result;
	else
		return result;

err:
	*endp = NULL;
	return 0;
}

/*
int main(void)
{
	const char* end;

	assert(strtoi("0", &end) == 0);
	assert(strtoi("1", &end) == 1);
	assert(strtoi("-1", &end) == -1);
	assert(strtoi("123", &end) == 123);
	assert(strtoi("-123", &end) == -123 && *end == 0);
	assert(strtoi("134523", &end) == 134523);
	assert(strtoi("-134523", &end) == -134523);
	assert(strtoi("11111", &end) == 11111);
	assert(strtoi("111", &end) == 111 && end && *end == '\0');

	assert(strtoi("assd", &end) && end == NULL);
	assert(strtoi("07423", &end) == 07423);
	assert(strtoi("-2147483648", &end) == -2147483648);
}
*/
